// Agent monitor in project infinite2.mas2j

/* Initial beliefs and rules */

timeout(4000). 

/****************************************************************/
/********** 				PROTOCOLS					*********/
/****************************************************************/

/* Brokering protocol, involves b, c, s */
global_type(brokering, Glob) :- // Esop 2012 paper by Pierre-Malo Denielou and Nobuko Yoshida, page 2
   Merge = choice([Off,Fork]) &
   Off=
   seq(sa(offer, 0),
      seq(sa(counter, 0), Merge)) &
   Fork= fork(seq(sa(final, 0),lambda),
              seq(sa(result, 0),lambda)) &
   Glob = seq(sa(item, 0),Merge). 

/* Alternating bit protocol with two parties and without constraints, involves alice_abp2, bob_abp2 TO BE DEFINED */
global_type(abp, S0) :- // Esop 2012 paper by Pierre-Malo Denielou and Nobuko Yoshida, page 5
	Msg1 = sa(mess(1), 0) &
	Msg2 = sa(mess(2), 0) &
	Ack1 = sa(ack(1), 0) &
	Ack2 = sa(ack(2), 0) &
	S0 = seq(Msg1, S1) & 
	S1 = choice([ seq(Ack1, S3), seq(Msg2, S2) ]) &
	S2 = choice([ seq(Ack1, S4), seq(Ack2, seq(Ack1,S0)) ]) &
	S3 = seq(Msg2, S4) &
	S4 = choice([ seq(Ack2, S0), seq(Msg1, S6) ]) &
	S6 = choice([ seq(Ack2, S1), seq(Ack1, seq(Ack2,S3)) ]).

/* Alternating bit protocol with two parties and with constraints, involves alice_abp2_constr, bob_abp2_constr TO BE DEFINED */	
global_type(abp2_constr, T) :- // Ancona, Mascardi, Barbieri FI 2013 
   M1M2=seq(sa(m1), seq(sa(m2), M1M2)) &
   M1A1=seq(sa(m1,1),seq(sa(a1,0),M1A1)) &
   M2A2=seq(sa(m2,1),seq(sa(a2,0),M2A2)) &
   T=fork(M1M2, fork(M1A1,M2A2)).

   
/* Alternating bit protocol with two parties, three messages and with constraints, involves alice_abp3_constr, bob_abp3_constr TO BE DEFINED */	
global_type(abp3_constr, T) :- // Ancona, Mascardi, Barbieri FI 2013 
   M1M2M3=seq(sa(m1),seq(sa(m2),seq(sa(m3),M1M2M3))) &
   M1A1=seq(sa(m1,1), seq(sa(a1,0), M1A1)) &
   M2A2=seq(sa(m2,1), seq(sa(a2,0), M2A2)) &
   M3A3=seq(sa(m3,1), seq(sa(a3,0), M3A3)) &
   T=fork(fork(M1M2M3,M1A1),fork(M2A2,M3A3)).

/* Infinite ping pong protocol, involves alice_pp, bob_pp */
global_type(ping_pong, Glob) :- // infinite ping pong
   Glob = seq(sa(ping, 0), 
      seq(sa(pong, 0), Glob)).  

/* simple a^n b^n, involves alice_abn, f */
global_type(abn, AB) :- // a^n b^n repeated once
	AB = seq(sa(a, 0), 
        choice([cat(AB, seq(sa(b, 0),lambda)), seq(sa(b, 0), lambda)])).

		
/* complex a^n b^n, involves alice_complex, bob_complex */
global_type(abn_complex, Forever) :- // a^n b^n repeated many times, with n varying each time (ICTCS 2012 paper), infinite
	Forever = cat(AB, Forever) & 
    AB = seq(sa(a, 0),
        choice([cat(AB, seq(sa(b, 0),lambda)), seq(sa(b, 0),lambda)])).			  
		

		
/****************************************************************/
/********** 		SENDING ACTION TYPES				*********/
/****************************************************************/


/* Types for the alternating bit protocol without constraints */
has_type(msg(alice_ab, bob_ab, tell, mess(N)), mess(N)).
has_type(msg(bob_ab, alice_ab, tell, ack(N)), ack(N)).

/* Types for the alternating bit protocol with constraints */
// to be defined
has_type(m1, m1).
has_type(m2, m2).
has_type(m3, m3).
has_type(a1, a1).
has_type(a2, a2).
has_type(a3, a3).

/* Types for the brokering protocol */	  
has_type(msg(b, c, tell, offer(Item, Offer)), offer) :- string(Item) & int(Offer).
has_type(msg(c, b, tell, counter(Item, Offer)), counter) :- string(Item) & int(Offer).
has_type(msg(b, s, tell, final(Res, Client, Item, Offer)), final) :- string(Res) & string(Client) & string(Item) & int(Offer).
has_type(msg(b, c, tell, result(Res, Item, Offer)), result) :- string(Res) & string(Item) & int(Offer).
has_type(msg(s, b, tell, item(Client, Item)), item) :- string(Client) & string(Item).

/* Types for the infinite ping pong protocol */
has_type(msg(alice_pp, bob_pp, tell, ping), ping).
has_type(msg(bob_pp, alice_pp, tell, pong), pong).

/* Types for the a^n b^n simple protocol */
has_type(msg(alice_abn, f, tell, X), X).


/* Types for the a^n b^n complex protocol */
has_type(msg(alice_complex, bob_complex, tell, a(X, T, N)), a) :- int(X) & int(T) & int(N).
has_type(msg(bob_complex, alice_complex, tell, b(X, N)), b) :- int(X) & int(N).

string(_).
int(_).
   

/****************************************************************/
/********** 				NEXT PREDICATE				*********/
/****************************************************************/

next(0,seq(sa(AType,N),T),AMessage,T,N) :- has_type(AMessage, AType).
next(M,seq(A,T1),A,T1,N) :- M > 0 & N = M - 1.
next(M,choice([T1,_]),A,T2,N) :- next(M,T1,A,T2,N). 
next(M,choice([_,T1]),A,T2,N) :- next(M,T1,A,T2,N).
next(N,fork(T1,T2),A,fork(T3,T2),M) :- next(N,T1,A,T3,M). 
next(N,fork(T1,T2),A,fork(T1,T3),M) :- next(N,T2,A,T3,M).
next(N,fork(T1,T2),A,fork(T3,T4),P) :- next(N,T1,A,T3,M) & M > 0 & next(M,T2,A,T4,P). 
next(N,fork(T1,T2),A,fork(T4,T3),P) :- next(N,T2,A,T3,M) & M > 0 & next(M,T1,A,T4,P). 
next(M,cat(T1,T2),A,cat(T3,T2),N) :- next(M,T1,A,T3,N).
next(M,cat(T1,T2),A,T3,N) :- empty(T1) & next(M,T2,A,T3,N).

/****************************************************************/
/********** 				EMPTY PREDICATE				*********/
/****************************************************************/

empty(lambda).
empty(choice([T1,_])) :- empty(T1).
empty(choice([_,T1])) :- empty(T1).
empty(fork(T1,T2)) :- empty(T1) & empty(T2).
empty(cat(T1,T2)) :- empty(T1) & empty(T2).

/****************************************************************/
/********** 			TYPE_CHECK PREDICATE			*********/
/****************************************************************/

type_check(msg(S, R, P, C), NewState) :- 
   current_state(LastState) &
   // .print("inside type_check: LastState = ", LastState , "\n") & 
   next(0, LastState, msg(S, R, P, C), NewState, 0).


/****************************************************************/
/********** 				CURRENT STATE				*********/
/****************************************************************/   
   
current_state(initial).

/****************************************************************/
/********** 				INITIAL GOAL				*********/
/****************************************************************/

!test(global_type(brokering)).
// !test(global_type(abp)).
// !test(global_type(abp2_constr)).
// !test(global_type(abp3_constr)).
// !test(global_type(ping_pong)).
// !test(global_type(abn)).
// !test(global_type(abn_complex)).

/* Plans */

/* Monitor's plans */

@test[atomic]
+!test(global_type(Name)) : current_state(initial) & global_type(Name, G) 
   <- !move_to_state(G).

@move2state[atomic]
+!move_to_state(NewState) : current_state(LastState) 
   <- -current_state(_);
      +current_state(NewState).	

@successfulMove[atomic]
+!type_check_message(msg(S, R, P, C)) : current_state(Current) & type_check(msg(S, R, P, C), NewState)  
   <- !move_to_state(NewState);
      .print("\nMessage ", msg(S, R, P, C), "\nleads from state ", Current, " to state ", NewState, "\n");
      .send(S, tell, ok_check(msg(S, R, P, C))).

@failingMoveAndProtocol
+!type_check_message(msg(S, R, P, C)) : current_state(Current) 
                                        & global_type(Name, _)
   <- .print("\n*** DYNAMIC TYPE-CHECKING ERROR ***\nMessage ", msg(S, R, P, C), " received within protocol ", Name, "\ncannot be accepted in the current state ", Current, "\n");
       !move_to_state(initial).

/**
 * Se vale current_state(initial) (cioe' si e' ancora nello stato iniziale, non
 * si e' ancora impostata la sessione) viene visualizzato un messaggio d'errore
 */
@messageReceptionKO
+msg(S, R, P, C): current_state(initial)
   <- .print("\n*** DYNAMIC TYPE-CHECKING ERROR ***\nMessage ", msg(S, R, P, C), " received but no protocol is ongoing successfully\n").
   
@messageReceptionOK
+msg(S, R, P, C)[source(S)]: timeout(MS)
   <- -msg(S, R, P, C)[source(S)]; 
      !type_check_message(msg(S, R, P, C));
	  !check_progress.
	  
+!check_progress : timeout(MS) 
   <- .wait({+msg(S1, R1, P1, C1)}, MS, Event); 
	  !aux_check_progress(Event).
	  
+!aux_check_progress(Event) : timeout(MS) 
                                           & Event < MS. 

+!aux_check_progress(Event) : timeout(MS) 
                                           & current_state(Current) 
										   & Event >= MS  
   <- .print("\n*** WARNING ***\nNo progress for ", Event, " milliseconds in the current state ", Current, "\n"); 
   !check_progress. 
